"""Email Message Reader."""
import base64
import email
from email.message import EmailMessage
from email.policy import default

from promail.core.attachements.attachments import Attachments


class Message:
    """Email Message reader."""

    def __init__(self, msg: dict) -> None:
        """Initalises Message object.

        Args:
            msg: email message data containing id
        """
        self.msg = email.message_from_bytes(msg["raw"])
        self._attachments = None

    @property
    def html_body(self) -> str:
        """Get HTML Body of email."""
        return str(
            self.msg.get_body(  # type: ignore
                preferencelist=("related", "html", "plain")
            )
        )

    @property
    def plain_text(self):
        """Get Plain text body of email."""
        return str(self.msg.get_body(preferencelist=("plain",)))  # type: ignore

    @property
    def sender(self) -> str:
        """Get sender of email."""
        return self.msg.get("from")

    @property
    def cc(self) -> str:
        """Get emails cc'd."""
        return self.msg.get("cc")

    @property
    def bcc(self) -> str:
        """Get emails ccc'd."""
        return self.msg.get("bcc")

    @property
    def message_id(self) -> str:
        """Get Message ID of email."""
        return self.msg.get("message-id")

    @property
    def to(self) -> str:
        """Get to field of email."""
        return self.msg.get("to")

    @property
    def subject(self) -> str:
        """Get Subject of email."""
        return self.msg.get("subject")

    @property
    def date(self):
        """Get Date of Email."""
        return self.msg.get("date")

    @property
    def attachments(self):
        """Get Email Attachments."""
        if self._attachments is None:
            self._attachments = {}
            for email_message_attachment in self.msg.iter_attachments():
                if email_message_attachment.is_attachment():
                    self._attachments[
                        email_message_attachment.get_filename()
                    ] = Attachments(email_message_attachment)
        return self._attachments

    def data(self) -> dict:
        """Get all Email fields."""
        return {
            "attachments": self.attachments,
            "date": self.date,
            "subject": self.subject,
            "to": self.to,
            "message_id": self.message_id,
            "bcc": self.bcc,
            "cc": self.cc,
            "html_body": self.html_body,
            "plain_text": self.plain_text,
        }

    def __str__(self) -> str:
        """String representation."""
        return self.subject or ""


class GmailMessage(Message):
    """Gmil Message object."""

    def __init__(self, msg: dict) -> None:
        """Initalises Message object.

        Args:
            msg: email message data containing id
        """
        super(GmailMessage, self).__init__(msg)
        # overwrite message with google specific settings
        self.msg = email.message_from_bytes(
            base64.urlsafe_b64decode(msg["raw"]), _class=EmailMessage, policy=default
        )



"""Email Filter implementation for Gmail."""
import re
from typing import Optional

from promail.filters.email_filter import EmailFilter


class GmailFilter(EmailFilter):
    """Email Filter Generates a query string used to query the email backend.

    Email filter is used by the email client to store
    which emails have been run with which filters.
    The Filter uses `name` and `version` to uniquely identify itself.
    Queries based on: https://seosly.com/gmail-search-operators/
    """

    TIME_FRAMES = {
        "d": "Day",
        "m": "Month",
        "y": "Year",
    }

    def _validate(self) -> None:
        fields = "newer_than", "older_than"
        for field in fields:
            value = getattr(self, f"_{field}")
            if value is None:
                continue
            value = value.strip().lower()
            if len(value) < 2:
                raise IndexError(
                    f"{value} is not valid for{field} is expected to be either None "
                    f"or a string representing the number of a time period. "
                    f"example: 6d for 6 days. valid options are: {self.TIME_FRAMES}"
                )

            elif value[-1] not in self.TIME_FRAMES.keys():
                raise ValueError(
                    f"{value} is not valid for{field},"
                    f"string must end in one of the following values: "
                    f"{list(self.TIME_FRAMES.keys())} "
                )
            elif not value[:-1].isdigit():
                raise ValueError(
                    f"{value} is not valid for{field}, "
                    f"expecting value to begin with a number"
                )

    @staticmethod
    def _join_tuple(term: str, data: Optional[tuple], seperator: str):
        if data is None:
            return ""
        if seperator == "OR":
            return " OR ".join([f"{term}:{d}" for d in data])
        elif seperator == " ":
            return f"{term}:(" + seperator.join([f"{d}" for d in data]) + ")"

    @staticmethod
    def _get_boalean(term: str, value: tuple, data: Optional[bool]):
        if data is None:
            return ""

        elif data:
            return f"{term}:{value[1]}"

        else:
            return f"{term}:{value[2]}"

    @staticmethod
    def _get_string(term: str, data: Optional[str]):
        if data is None:
            return ""
        return f"{term}:{data}"

    @staticmethod
    def _get_date(term, date):
        if date is None:
            return ""
        return f"{term}:{date.strftime('%G/%m/%d')}"

    @staticmethod
    def _get_toggle(term, data):
        """Format toggle."""
        if data:
            return f"{term}:{data}"
        return ""

    @property
    def sender(self):
        """Search query based on sender."""
        return self._join_tuple("from", self._sender, seperator="OR")

    @property
    def read(self):
        """Search query based on read."""
        return self._get_boalean("read", ["read", "unread"], self._read)

    @property
    def newer_than(self):
        """Search query based on newer_than."""
        return self._get_string("newer_than", self._newer_than)

    @property
    def not_sender(self):
        """Search query based on not_sender."""
        return self._join_tuple("NOT from", self._not_sender, seperator="OR")

    @property
    def older_than(self):
        """Search query based on older_than."""
        return self._get_string("older_than", self._older_than)

    @property
    def phrase(self):
        """Search query based on phrase."""
        if self._phrase is None:
            return ""
        return self._join_tuple(
            "", tuple([f'"{phrase}"' for phrase in self._phrase]), seperator="AND"
        )

    @property
    def size(self):
        """Search query based on size."""
        return "" if self._size is None else f"size:{self._size}"

    @property
    def size_larger(self):
        """Search query based on size_larger."""
        return self._get_string("larger", self._size_larger)

    @property
    def size_smaller(self):
        """Search query based on size_smaller."""
        return self._get_string("larger", self._size_smaller)

    @property
    def sent_after(self):
        """Search query based on sent_after."""
        return self._get_date("after", self._sent_after)

    @property
    def sent_before(self):
        """Search query based on sent_before."""
        return self._get_date("before", self._sent_before)

    @property
    def starred(self):
        """Search query based on starred."""
        data = True if self._starred else None
        return self._get_toggle("starred", data)

    @property
    def keyword(self):
        """Search query based on keyword."""
        if self._keyword is None:
            return ""
        else:
            return "(" + " ".join(self._keyword) + ")"

    @property
    def important(self):
        """Search query based on important."""
        if self._important is None:
            return ""
        return self._get_toggle("is", "important")

    @property
    def filename(self):
        """Search query based on filename."""
        return self._join_tuple("filename", self._filename, "OR")

    @property
    def category(self):
        """Search query based on category."""
        return self._get_string("category", self._category)

    @property
    def bcc(self):
        """Search query based on bcc."""
        return self._join_tuple("bcc", self._bcc, "AND")

    @property
    def around(self):
        """Search query based on around."""
        if self._around is None:
            return ""
        return (
            f"{self._around['first_term']}"
            "AROUND {self._around['apart']} "
            "{self._around['second_term']}"
        )

    @property
    def attachment(self):
        """Search query based on attachment."""
        return self._get_toggle("has", "attachment" if self._attachment else None)

    @property
    def label(self):
        """Search query based on label."""
        return self._get_string("label", self._label)

    @property
    def folder(self):
        """Search query based on folder."""
        return self._get_string("folder", self._folder)

    @property
    def cc(self):
        """Search query based on cc."""
        return self._join_tuple("cc", self._cc, "AND")

    @property
    def to(self):
        """Search query based on to."""
        return self._join_tuple("to", self._to, "AND")

    @property
    def subject(self):
        """Search query based on subject."""
        return self._get_string("subject", self._subject)

    def get_filter_string(self) -> str:
        """Creates string to query email based on parameters."""
        return re.sub(
            " +",
            " ",
            " ".join(
                (
                    self.around,
                    self.attachment,
                    self.bcc,
                    self.category,
                    self.cc,
                    self.filename,
                    self.folder,
                    self.important,
                    self.keyword,
                    self.label,
                    self.newer_than,
                    self.not_sender,
                    self.older_than,
                    self.phrase,
                    self.read,
                    self.sender,
                    self.sent_after,
                    self.sent_before,
                    self.size,
                    self.size_larger,
                    self.size_smaller,
                    self.starred,
                    self.to,
                    self.subject,
                )
            ).strip(),
        )

    def filter_results(self, messages):
        """Removes messages in self.processed."""
        return filter(lambda msg: msg["id"] not in self.processed, messages)


# a = GmailFilter(
#     load_processed=True,
#     sender=("Antoinewood@gmail.com", "sue"),
#     older_than="99d",
#     sent_after=datetime.now(),
# )
#
# print(a.get_filter_string())
# # l = list(set(a.__dict__.keys()))
# # l.sort()
# # for value in l:
# #     print(
# #         f"self.{value.strip('_')}", end = ", "
# #     )



"""Email Attachments."""
import email
from email.message import EmailMessage
from io import BytesIO


class Attachments:
    """Email Attachment Reader."""

    manager = email.contentmanager.raw_data_manager

    def __init__(self, email_attachment: EmailMessage):
        """Initializes Email Attachment."""
        self.email_attachment = email_attachment

    @property
    def filename(self):
        """Get filename."""
        return self.email_attachment.get_filename()

    def save_file(self, path):
        """Saves file to path provided."""
        with open(path, "wb") as file:
            file.write(self.email_attachment.get_content(content_manager=self.manager))

    def get_data(self) -> BytesIO:
        """Get file data as an inmemory as a BytesIO file-like object."""
        obj = BytesIO()
        obj.write(self.email_attachment.get_content(content_manager=self.manager))
        obj.seek(0)
        return obj

    def __str__(self):
        """Get string representation."""
        return self.filename

    def __repr__(self):
        """Get repr representation."""
        return f"Attachments({self.__str__()})"



"""EmbeddedAttachments."""

from email.utils import make_msgid


class EmbeddedAttachments:
    """EmbeddedAttachments in Email.

    EmbeddedAttachments are used to include your local files
    in the body of an email.
    """

    def __init__(self, filepath):
        """Initializes EmbeddedAttachments."""
        self.filepath = filepath
        self._cid = make_msgid(domain="promail")

    @property
    def cid(self):
        """Returns unique identifier for reference in email."""
        return self._cid

    def __str__(self):
        """Returns CID stripped of brackets to be used as  src={cid}."""
        return self._cid[1:-1]

    def read(self):
        """Returns Byte info of file."""
        with open(self.filepath, "rb") as file:
            return file.read()



"""Email Clients."""
from .gmail import GmailClient
from .imap_client import ImapClient
from .microsoft import HotmailClient
from .smtp_client import SmtpClient


try:
    from importlib.metadata import version, PackageNotFoundError  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

__all__ = [
    "email_manager",
    "gmail",
    "GmailClient",
    "HotmailClient",
    "ImapClient",
    "SmtpClient",
]



"""Abstract Classes for Inbound and outbound mail."""
import abc
import mimetypes
import os
from email.message import EmailMessage
from typing import Callable, List, Optional

from promail.core.embedded_attachments import EmbeddedAttachments
from promail.core.messages.messages import Message
from promail.filters.email_filter import EmailFilter


class EmailManager:
    """Super class inherited by OutBoundInbound manager."""

    def __init__(self, account):
        """Initializes email manager."""
        self._account = account


class OutBoundManager(abc.ABC, EmailManager):
    """Outbound Mail class template."""

    def __init__(self, account, *args, **kwargs):
        """Initializes OutBoundManager."""
        super(OutBoundManager, self).__init__(account, *args, **kwargs)

    def send_email(
        self,
        recipients: str = "",
        cc: str = "",
        bcc: str = "",
        subject: str = "",
        htmltext: str = "",
        plaintext: str = "",
        embedded_attachments: Optional[List[EmbeddedAttachments]] = None,
        attachements: Optional[list] = None,
    ) -> None:
        """Send an email."""
        pass

    @staticmethod
    def guess_types(path: str) -> tuple:
        """Will attempt to guess ctype, subtype and maintype of file.

        Based on https://docs.python.org/3/library/email.examples.html

        Args:
            path: path to attachment file

        Returns:
            tuple: maintype, subtype of file
        """
        ctype, encoding = mimetypes.guess_type(path)
        if ctype is None or encoding is not None:
            # No guess could be made, or the file is encoded (compressed), so
            # use a generic bag-of-bits type.
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)
        return maintype, subtype

    def add_attachments(self, msg: EmailMessage, attachments: list) -> None:
        """Add attachment to email."""
        for path in attachments:
            filename = os.path.basename(path)
            maintype, subtype = self.guess_types(path)
            with open(path, "rb") as fp:
                msg.add_attachment(
                    fp.read(), maintype=maintype, subtype=subtype, filename=filename
                )

    def create_message(
        self,
        recipients: str = "",
        cc: str = "",
        bcc: str = "",
        subject: str = "",
        htmltext: str = "",
        plaintext: str = "",
        embedded_attachments: Optional[List[EmbeddedAttachments]] = None,
        attachements: Optional[list] = None,
    ) -> EmailMessage:
        """Create Email Message."""
        if attachements is None:
            attachements = []
        if embedded_attachments is None:
            embedded_attachments = []
        msg = EmailMessage()
        msg["Subject"] = subject
        msg["From"] = self._account
        msg["To"] = recipients
        msg["Cc"] = cc
        msg["Bcc"] = bcc
        msg.set_content(plaintext)
        msg.add_alternative(htmltext, subtype="html")
        for embedded_attachment in embedded_attachments:
            maintype, subtype = self.guess_types(embedded_attachment.filepath)
            msg.get_payload()[1].add_related(
                embedded_attachment.read(),
                maintype,
                subtype,
                cid=embedded_attachment.cid,
            )

        self.add_attachments(msg, attachements)
        return msg


class InBoundManager(abc.ABC, EmailManager):
    """InBound Mail class template."""

    def __init__(self, account):
        """Initializes Inbound Email manager."""
        self._registered_functions: dict = {}
        self._last_email = None
        self._filter_class = None
        super(InBoundManager, self).__init__(account)

    @property
    def registered_functions(self) -> dict:
        """Get Dictionary of (key) filters and (value) registered functions."""
        return self._registered_functions

    def retrieve_last_items(self: object, max_items: int = 100) -> List[Message]:
        """Get a list of last n items received in inbox.

        Args:
            max_items: The Maximum number of items to return

        Raises:
            NotImplementedError: If not implemented by subclass.
        """
        raise NotImplementedError(__name__ + " not Implemented")

    def _process_filter_messages(
        self,
        email_filter: EmailFilter,
        page_size: int = 100,
        page_token: Optional[str] = None,
    ):
        """Queries Email Server for new messages.

         That match filter requisites passes each matched message
         to their registered functions.

        Args:
            email_filter: Email Filter Object, must be a key
                of self._registered_functions.
            page_size: Number of emails to pull per query,
                max number platform dependent (GMAIL: 500).
            page_token: Pagenation Token
                (may not be used on all platforms).

        Raises:
            NotImplementedError: If not implemented by subclass.
        """
        raise NotImplementedError

    def process(self, page_size: int = 100) -> None:
        """Process all filters."""
        for email_filter in self._registered_functions:
            self._process_filter_messages(
                email_filter, page_size=page_size, page_token=None
            )

    def register(self, **filter_args):
        """Registers a listener function."""

        def decorator(func: Callable):
            def wrapper(email: EmailMessage):
                func(email)

            f = self._filter_class(**filter_args)

            if f not in self._registered_functions:
                self._registered_functions[f] = set()
            self._registered_functions[f].add(wrapper)

        return decorator



"""Imap filter."""
from typing import Optional

from promail.filters.email_filter import EmailFilter


class ImapFilter(EmailFilter):
    """Imap Filter."""

    @staticmethod
    def _get_date(term, date):
        if date is None:
            return ""
        return [term, date.strftime("%G/%m/%d")]

    @staticmethod
    def _get_string(term: str, data: Optional[str]):
        if data is None:
            return None
        return [term, data]

    @staticmethod
    def _join_tuple(term: str, data: Optional[tuple]):
        if data is None or len(data) == 0:
            q = None
        elif len(data) == 1:
            q = [term, data[0]]
        else:
            q = ["OR"] * len(data)
            for d in data:
                q.extend([term, d])
        return q

    @property
    def bcc(self) -> Optional[list]:
        """Email BCC Query."""
        return self._join_tuple("BCC", self._bcc)

    @property
    def body(self) -> Optional[list]:
        """Messages that contain the specified string in the body of the message."""
        return self._get_string("BODY", self._body)

    @property
    def cc(self) -> Optional[list]:
        """CC filter.

        Messages that contain the specified string
        in the envelope structure's CC field.

        Returns:
            list: ["CC", "emails]"
        """
        return self._join_tuple("CC", self._cc)

    @property
    def sender(self):
        """Search query based on sender."""
        return self._get_string("FROM", self._sender)

    @property
    def sent_before(self):
        """Messages whose internal date is earlier than the specified date."""
        return self._get_string("BEFORE ", self._sent_before)

    @property
    def size_larger(self):
        """Search query based on size_larger."""
        return self._get_string("LARGER", self._size_larger)

    @property
    def subject(self) -> Optional[list]:
        """Email Subject Query."""
        return self._get_string("SUBJECT", self._subject)

    @property
    def to(self) -> Optional[list]:
        """Email Subject Query."""
        return self._join_tuple("TO", self._to)

    def get_filter_list(self) -> list:
        """Combine search parameters into query."""
        f_list = []
        data = (
            self.subject,
            self.bcc,
        )
        for parameter in data:
            if parameter is not None:
                f_list.extend(parameter)
        return f_list

    def filter_results(self, messages):
        """Removes messages in self.processed."""
        return filter(lambda msg_id: msg_id not in self.processed, messages)



"""Outlook/Hotmail Mail Client."""
from promail.clients.imap_client import ImapClient
from promail.clients.smtp_client import SmtpClient


class HotmailClient(SmtpClient, ImapClient):
    """Hotmail/Outlook.com Client.

    Should work on most SMTP/IMAP based if correct hosts are set.
    """

    def __init__(
        self,
        account,
        password,
        smtp_host="smtp-mail.outlook.com",
        imap_host="outlook.office365.com",
    ):
        """Hotmail Email Client Initiation."""
        SmtpClient.__init__(self, account=account, password=password, host=smtp_host)
        ImapClient.__init__(self, account=account, password=password, host=imap_host)



"""Imap Client."""
import imaplib
import logging

from imapclient import IMAPClient  # type: ignore

from promail.clients.email_manager import InBoundManager
from promail.core.messages.messages import Message
from promail.filters.email_filter import EmailFilter
from promail.filters.imap_filter import ImapFilter


class ImapClient(InBoundManager):
    """Imap Client."""

    def __init__(self, account, password, host, **kwargs):
        """Initiate IMAP client."""
        super(ImapClient, self).__init__(account)
        self._imap_host = host
        self._password = password
        self._filter_class = ImapFilter

    def _process_filter_messages(
        self, email_filter: EmailFilter, page_size=100, page_token=None
    ):
        # get emails that match search criteria
        with IMAPClient(host=self._imap_host) as client:
            client.login(self._account, self._password)
            client.select_folder("INBOX")
            results = client.search(email_filter.get_filter_list())

            # remove emails that have already been processed
            messages = email_filter.filter_results(results)
            try:
                for uid, message_data in client.fetch(messages, "RFC822").items():
                    email_message = Message({"raw": message_data[b"RFC822"]})
                    for func in self._registered_functions[email_filter]:
                        func(email_message)
                        email_filter.add_processed(uid)
            except imaplib.IMAP4.error as e:
                # FETCH command error if messages len == 0
                logging.info(e)



"""Email Filter."""

import abc
import hashlib
import os.path
import pickle
from datetime import datetime
from typing import Optional


class EmailFilter(abc.ABC):
    """Email Filter Generates a query string used to query the email backend.

    Email filter is used by the email client to store
    which emails have been run with which filters.
    The Filter uses `name` and `version` to uniquely identify itself.
    Queries based on: https://seosly.com/gmail-search-operators/
    """

    def __init__(
        self,
        name: str,
        run_completed: bool = False,
        around: Optional[dict] = None,
        attachment: Optional[bool] = None,
        bcc: Optional[tuple] = None,
        body: Optional[str] = None,
        category: Optional[str] = None,
        cc: Optional[tuple] = None,
        filename: Optional[tuple] = None,
        folder: Optional[tuple] = None,
        important: Optional[bool] = None,
        label: Optional[tuple] = None,
        keyword: Optional[tuple] = None,
        newer_than: Optional[str] = None,
        not_sender: Optional[tuple] = None,
        older_than: Optional[str] = None,
        phrase: Optional[tuple] = None,
        sender: Optional[tuple] = None,
        size: Optional[int] = None,
        size_larger: Optional[str] = None,
        size_smaller: Optional[str] = None,
        sent_after: Optional[datetime] = None,
        sent_before: Optional[datetime] = None,
        starred: Optional[bool] = None,
        subject: Optional[str] = None,
        to: Optional[tuple] = None,
        read: Optional[bool] = None,
        version: Optional[str] = None,
    ):
        """Initializes email filter.

        Args:
            name: User defined name of filter.
                The name along with version is used to identify
                the filter internally to check which messages
                have been already run with a particular filter.
            run_completed: If True it will include messages
                that have already been processed.
            around: Will check body text for first term,
                within apart words of second term
                requires dictionary in form of
                {"first_term": "Term", "apart": 20, "second_term": "Other"}
            attachment: True will filter only emails with attachments,
                False without attachments, None will show all.
            bcc: Will look for emails that include terms in bcc,
                expects tuple of strings if more than one term
                provided will search based on "OR"
            body: Messages that contain the specified string in the body of the message
            category: filter on category
            cc: filter on cc
            filename: filter on filename
            folder: filter on folder
            important: filter on if important
            label: filter on label
            keyword: filter on keyword
            newer_than: filter on newer than
            not_sender: filter our senders
            older_than: filter older than
            phrase: filter by including phrase
            sender: filter on sender
            size: Minimum size of email in bytes
            size_larger: filter on size
            size_smaller: filter size
            sent_after: filter on date sent
            sent_before: filter on date sent
            starred: filter if starred
            subject: filter on phrase in subject
            to: filter on to fieled
            read: filter on if read
            version: Version will control whether
        """
        self.name = name
        self._newer_than = newer_than
        self._not_sender = not_sender
        self._older_than = older_than
        self._phrase = phrase
        self._size = size
        self._size_larger = size_larger
        self._size_smaller = size_smaller
        self._sent_after = sent_after
        self._sent_before = sent_before
        self._starred = starred
        self._keyword = keyword
        self._important = important
        self._filename = filename
        self._category = category
        self._bcc = bcc
        self._around = around
        self._attachment = attachment
        self._label = label
        self._folder = folder
        self._body = body
        self._cc = cc
        self._to = to
        self._read = read
        self._sender = sender
        self._version = version
        self._subject = subject
        self.save_folder = "processed_emails"

        self._validate()
        if run_completed:
            self.processed = set()
        else:
            self.processed = self.load_processed_ids()

    def _validate(self):
        """Validates inputs."""
        pass

    def filter_results(self, messages):
        """Removes messages in self.processed."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def processed_filename(self):
        """Relative path to pickle file."""
        return f"{self.save_folder}/{hash(self)}.bin"

    def __hash__(self) -> int:
        """Hash based on name, version."""
        fields = (
            self.name,
            self._version,
        )
        return int(hashlib.md5(str(fields).encode()).hexdigest(), 16)

    def load_processed_ids(self):
        """Loads A Set of ids that have been processed with this filter."""
        try:
            with open(self.processed_filename, "rb") as file:
                return pickle.load(file)
        except FileNotFoundError:
            return set()

    def add_processed(self, email_id: str) -> None:
        """Add Message to list of processed Messages."""
        self.processed.add(email_id)
        if not os.path.exists(self.save_folder):
            os.makedirs(self.save_folder)
        with open(self.processed_filename, "wb") as file:
            pickle.dump(self.processed, file)

    def get_filter_string(self) -> str:
        """Creates string to query email based on parameters.

        Returns: Query String.

        Raises:
            NotImplementedError: method not implemented.
        """
        raise NotImplementedError(__name__ + " not implemented")

    def get_filter_list(self) -> list:
        """Creates list to query email based on parameters.

        Returns: Query String.

        Raises:
            NotImplementedError: method not implemented.
        """
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def body(self):
        """Messages that contain the specified string in the body of the message."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def sender(self):
        """Search query based on sender."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def read(self):
        """Search query based on read."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def newer_than(self):
        """Search query based on newer_than."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def not_sender(self):
        """Search query based on not_sender."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def older_than(self):
        """Search query based on older_than."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def phrase(self):
        """Search query based on phrase."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def size(self):
        """Search query based on size."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def size_larger(self):
        """Search query based on size_larger."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def size_smaller(self):
        """Search query based on size_smaller."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def sent_after(self):
        """Search query based on sent_after."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def sent_before(self):
        """Search query based on sent_before."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def starred(self):
        """Search query based on starred."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def keyword(self):
        """Search query based on keyword."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def important(self):
        """Search query based on important."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def filename(self):
        """Search query based on filename."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def category(self):
        """Search query based on category."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def bcc(self):
        """Search query based on bcc."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def around(self):
        """Search query based on around."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def attachment(self):
        """Search query based on attachment."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def label(self):
        """Search query based on label."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def folder(self):
        """Search query based on folder."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def cc(self):
        """Search query based on cc."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def to(self):
        """Search query based on to."""
        raise NotImplementedError(__name__ + " not implemented")

    @property
    def subject(self):
        """Search query based on subject."""
        raise NotImplementedError(__name__ + " not implemented")



"""SMPT Client."""
import smtplib
from typing import List, Optional

from promail.clients.email_manager import OutBoundManager
from promail.core.embedded_attachments import EmbeddedAttachments


class SmtpClient(OutBoundManager):
    """SMPT Client."""

    def __init__(self, account: str, password: str, host: str, port: int = 587) -> None:
        """Initiates SMPT Client.

        Args:
            account (str): Email Address
            password (str): password
            host (str): url of host.
            port (int): Port to connect on.
        """
        super(SmtpClient, self).__init__(
            account=account, password=password, host=host, port=port
        )
        self._password = password
        self._smtp_host = host
        self._port = port

    def login(self, server):
        """Login to server."""
        server.ehlo()
        server.starttls()
        server.login(self._account, self._password)

    def send_email(
        self,
        recipients: str = "",
        cc: str = "",
        bcc: str = "",
        subject: str = "",
        htmltext: str = "",
        plaintext: str = "",
        embedded_attachments: Optional[List[EmbeddedAttachments]] = None,
        attachements: Optional[list] = None,
    ) -> None:
        """Send Email.

        Args:
            recipients: semicolon seperated string of recipients will show in to section
            cc: semicolon seperated string of recipients will  be cc'd on email
            bcc: semicolon seperated string of recipients will be bcc'd on email
            subject: Subject of the email
            htmltext: An HTML string of text
            plaintext: A plain text alternative text
            embedded_attachments: List of EmbeddedAttachments included in email.
            attachements: list of filenames or filelike objects
        """
        msg = self.create_message(
            recipients,
            cc,
            bcc,
            subject,
            htmltext,
            plaintext,
            embedded_attachments,
            attachements,
        )
        with smtplib.SMTP(self._smtp_host, self._port) as server:
            self.login(server)
            server.sendmail(self._account, recipients, msg.as_string())
            server.close()



"""Gmail Mail Client."""
import base64
import logging
import os.path
from typing import List, Optional

from google.auth.transport.requests import Request  # type: ignore
from google.oauth2.credentials import Credentials  # type: ignore

from google_auth_oauthlib.flow import InstalledAppFlow  # type: ignore

from googleapiclient.discovery import build  # type: ignore
from googleapiclient.errors import HttpError  # type: ignore

from promail.clients.email_manager import (
    InBoundManager,
    OutBoundManager,
)
from promail.core.embedded_attachments import EmbeddedAttachments
from promail.core.messages.messages import GmailMessage
from promail.filters.gmail_filter import GmailFilter

logging.basicConfig(level=logging.INFO)


class GmailClient(OutBoundManager, InBoundManager):
    """Gmail Client."""

    SCOPES = [
        "https://www.googleapis.com/auth/gmail.readonly",
        "https://www.googleapis.com/auth/gmail.send",
    ]

    def __init__(
        self,
        account: str,
        token_path: str = "",
        credentials: str = "",
        clear_token: bool = False,
    ) -> None:  # ignore
        """Initiates Gmail Client.

        Args:
            clear_token: Should class clear access token on exit
            account: Gmail Email Address
            token_path: Path of Gmail token. Leave blank for default
            credentials: Location of Gmail API Credentials.

                Leave Blank to use promail's credentials.
                Promail is shipped with credentials, however there
                is a daily maximum of 1 Billion Actions
                shared amongst all users of the software.

                To keep this pool available we recommend that for heavy or
                production uses you create your own key, so that
                we don't exhaust current resources.
                The Gmail API is free so there are no cost implications.


        """
        super(GmailClient, self).__init__(account)
        self._filter_class = GmailFilter
        sanitized_account: str = "".join(x for x in account if x.isalnum())
        self._token_path: str = token_path or os.path.join(
            os.getcwd(),
            ".credentials",
            "gmail",
            f"{sanitized_account}.json",
        )
        self.gmail_credentials: str = credentials or os.environ.get(
            "GMAIL_CREDENTIALS", ""
        )
        self.service = self.login()
        self._clear_token = clear_token

    def login(self):
        """Logs into Gmail using OAuth and saves a token for later use."""
        creds = None
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        if os.path.exists(self._token_path):
            creds = Credentials.from_authorized_user_file(
                self._token_path, GmailClient.SCOPES
            )
        # If there are no (valid) credentials available, let the user log in.

        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                try:
                    flow = InstalledAppFlow.from_client_secrets_file(
                        self.gmail_credentials, GmailClient.SCOPES
                    )
                except FileNotFoundError as e:
                    logging.error(e)
                    logging.error(
                        "Please include the location of your "
                        "GMAIL credentials json in the class initiatiation"
                        " or Set GMAIL_CREDENTIALS env variable to the "
                        "location of your gmail credentials json."
                    )
                    exit(400)

                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run

            # folder doesn't exist create it
            if not os.path.exists(os.path.dirname(self._token_path)):
                os.makedirs(os.path.dirname(self._token_path))

            with open(self._token_path, "w") as token:
                token.write(creds.to_json())

        self.service = build("gmail", "v1", credentials=creds)
        return self.service

    def send_email(
        self,
        recipients: str = "",
        cc: str = "",
        bcc: str = "",
        subject: str = "",
        htmltext: str = "",
        plaintext: str = "",
        embedded_attachments: Optional[List[EmbeddedAttachments]] = None,
        attachements: Optional[list] = None,
    ) -> None:
        """Send email."""
        msg = self.create_message(
            recipients,
            cc,
            bcc,
            subject,
            htmltext,
            plaintext,
            embedded_attachments,
            attachements,
        )

        raw_data = base64.urlsafe_b64encode(msg.as_bytes())
        raw = raw_data.decode()
        try:
            # by specifying "me" it will always send from logged-in user
            message = (
                self.service.users()
                .messages()
                .send(userId="me", body={"raw": raw})
                .execute()
            )
            return message
        except HttpError as error:
            print("An error occurred: %s" % error)

    # inbound

    def _process_filter_messages(self, email_filter, page_size=100, page_token=None):
        # get emails that match search criteria
        results = (
            self.service.users()
            .messages()
            .list(
                userId="me",
                maxResults=page_size,
                q=email_filter.get_filter_string(),
                pageToken=page_token,
            )
            .execute()
        )

        # remove emails that have already been processed
        messages = email_filter.filter_results(results.get("messages", []))

        for message in messages:
            current_message = (
                self.service.users()
                .messages()
                .get(userId="me", id=message["id"], format="raw", metadataHeaders=None)
                .execute()
            )
            email_message = GmailMessage(current_message)
            # pass email to each functioned registered with filter
            for func in self._registered_functions[email_filter]:
                func(email_message)
                email_filter.add_processed(message["id"])

        # process next page in paginated results.
        next_page = results.get("nextPageToken")
        if next_page:
            self._process_filter_messages(
                email_filter, page_size=100, page_token=next_page
            )

    # gmail specific functionality
    def mailboxes(self):
        """Labels that we can filter by."""
        return [
            label["id"]
            for label in self.service.users()
            .labels()
            .list(userId="me")
            .execute()["labels"]
        ]

    def __exit__(self, exc_type, exc_val, exc_tb):
        """If clear token flag has been set will delete token on exit."""
        if self._clear_token:
            os.remove(self._token_path)


#
# client = GmailClient("antoinewood@gmail.com")
#
#
# @client.register(
#     name="search",
#     sender=("antoine",),
#     newer_than="4d",
#     version="10",
#     run_completed=True,
# )
# def print_subjects1(email):
#     if email.cc:
#         print(dir(email.cc))
#         print("2", type(email.cc), email.attachments)
#
#
# client.process(100)
# #
# # @client.register(name="search", sender=("antoine",), newer_than="100d")
# # def print_subjects2(email):
# #     print("1", email.subject)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['promail',
 'promail.clients',
 'promail.core',
 'promail.core.attachements',
 'promail.core.messages',
 'promail.filters']

package_data = \
{'': ['*']}

install_requires = \
['IMAPClient>=2.3.1,<3.0.0',
 'google-api-core>=2.8.0,<3.0.0',
 'google-api-python-client>=2.47.0,<3.0.0',
 'google-auth-httplib2>=0.1.0,<0.2.0',
 'google-auth-oauthlib>=0.5.1,<0.6.0',
 'google-auth>=2.6.6,<3.0.0',
 'nox>=2022.1.7,<2023.0.0',
 'python-dotenv>=0.1.0,<0.2.0',
 'requests>=2.27.1,<3.0.0',
 'tabulate>=0.8.9,<0.9.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.11.3,<5.0.0']}

setup_kwargs = {
    'name': 'promail',
    'version': '0.8.0',
    'description': 'Promail: The Python Email Automation Framework',
    'long_description': '[![Tests](https://github.com/trafire/promail/workflows/Tests/badge.svg)](https://github.com/trafire/promail/actions?workflow=Tests)\n[![Codecov](https://codecov.io/gh/trafire/promail/branch/main/graph/badge.svg)](https://codecov.io/gh/trafire/promail)\n[![PyPI](https://img.shields.io/pypi/v/promail.svg)](https://pypi.org/project/promail/)\n[![Read the Docs](https://readthedocs.org/projects/promail/badge/)](https://promail.readthedocs.io/)\n# Promail\n\nPromail along with its sister library Promail-Templates aims \nto close the email gap between what you as an individual can make your\nemail do with little effort and what enterprise users do\n\n- Automated Professional Rich content HTML emails\n- Allow you to write your own pluggins that do arbitrary things depending on the content of the email.\n\n## Installation\n```\npip install promail\n```\n## Simple Usage\n\n```python\nfrom promail.clients.gmail import GmailClient\n\nclient = GmailClient("your-gmail@gmail.com")\n# The first time you do this it will open a web browser allowing you to sign into your google account directly\nclient.send_email()\n\n```\n',
    'author': 'Antoine Wood',
    'author_email': 'antoinewood@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/trafire/promail',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
